/**
   @author Shin'ichiro Nakaoka
*/

#include "SimulationBar.h"
#include "SimulatorItem.h"
#include "WorldItem.h"
#include <cnoid/TimeBar>
#include <cnoid/RootItem>
#include <cnoid/ItemTreeView>
#include <cnoid/MessageView>
#include <boost/bind.hpp>
#include "gettext.h"

using namespace std;
using namespace boost;
using namespace cnoid;


SimulationBar* SimulationBar::initialize(ExtensionManager* ext)
{
    return instance();
}


SimulationBar* SimulationBar::instance()
{
    static SimulationBar* instance = new SimulationBar();
    return instance;
}


SimulationBar::SimulationBar()
    : ToolBar(N_("SimulationBar")),
      os(MessageView::mainInstance()->cout())
{
    addButton(QIcon(":/Body/icons/store-world-initial.png"),
              _("Store body positions to the initial world state"))->
        sigClicked().connect(bind(&SimulationBar::onStoreInitialClicked, this));
    
    addButton(QIcon(":/Body/icons/restore-world-initial.png"),
              _("Restore body positions from the initial world state"))->
        sigClicked().connect(bind(&SimulationBar::onRestoreInitialClicked, this));

    typedef boost::function<void(SimulatorItem* simulator)> Callback;

    addButton(QIcon(":/Body/icons/start-simulation.png"), _("Start simulation from the beginning"))->
        sigClicked().connect(bind(&SimulationBar::forEachSimulator, this,
                                  Callback(bind(&SimulationBar::startSimulation, this, _1, true))));

    addButton(QIcon(":/Body/icons/restart-simulation.png"),
              _("Start simulation from the current state"))->
        sigClicked().connect(bind(&SimulationBar::forEachSimulator, this,
                                  Callback(bind(&SimulationBar::startSimulation, this, _1, false))));
    
    addButton(QIcon(":/Body/icons/stop-simulation.png"), _("Stop simulation"))->
        sigClicked().connect(bind(&SimulationBar::onStopSimulationClicked, this));
}


SimulationBar::~SimulationBar()
{

}


static void forEachTargetBodyItem(function<void(BodyItem*)> callback)
{
    ItemTreeView* itemTreeView = ItemTreeView::instance();

    ItemList<BodyItem> bodyItems = RootItem::instance()->getSubItems<BodyItem>();
    for(int i=0; i < bodyItems.size(); ++i){
        BodyItem* bodyItem = bodyItems[i];
        bool isTarget = itemTreeView->isItemSelected(bodyItem);
        if(!isTarget){
            WorldItem* worldItem = bodyItem->findOwnerItem<WorldItem>();
            if(worldItem && itemTreeView->isItemSelected(worldItem)){
                isTarget = true;
            }
        }
        if(isTarget){
            callback(bodyItem);
        }
    }
}


static void storeInitialBodyState(BodyItem* bodyItem)
{
    bodyItem->storeInitialState();
    MessageView::instance()->put(fmt(_("Current state of %1% has been set to the initial state.\n"))
                                 % bodyItem->name());
}


void SimulationBar::onStoreInitialClicked()
{
    forEachTargetBodyItem(storeInitialBodyState);
    os.flush();
}


void SimulationBar::onRestoreInitialClicked()
{
    forEachTargetBodyItem(function<void(BodyItem*)>(bind(&BodyItem::restoreInitialState, _1)));
}


void SimulationBar::forEachSimulator(function<void(SimulatorItem* simulator)> callback)
{
    /*
    ItemList<SimulatorItem> simulators =
        ItemTreeView::mainInstance()->selectedItems<SimulatorItem>();
    */
    ItemList<SimulatorItem> simulators =
        ItemTreeView::mainInstance()->selectedItems<SimulatorItem>();

    typedef map<WorldItem*, SimulatorItem*> WorldToSimulatorMap;
    WorldToSimulatorMap worldToSimulator;

    for(int i=0; i < simulators.size(); ++i){
        SimulatorItem* simulator = simulators[i];
        WorldItem* world = simulator->findOwnerItem<WorldItem>();
        if(world){
            WorldToSimulatorMap::iterator p = worldToSimulator.find(world);
            if(p == worldToSimulator.end()){
                worldToSimulator[world] = simulator;
            } else {
                p->second = 0; // skip if multiple simulators are selected
            }
        }
    }

    for(int i=0; i < simulators.size(); ++i){
        SimulatorItem* simulator = simulators[i];
        WorldItem* world = simulator->findOwnerItem<WorldItem>();
        if(!world){
            os << (fmt(_("%1% cannot be processed because it is not related with a world."))
                   % simulator->name()) << endl;
        } else {
            WorldToSimulatorMap::iterator p = worldToSimulator.find(world);
            if(p != worldToSimulator.end()){
                if(!p->second){
                    os << (fmt(_("%1% cannot be processed because another simulator"
                                 "in the same world is also selected."))
                           % simulator->name()) << endl;
                } else {
                    callback(simulator);
                }
            }
        }
    }
}


void SimulationBar::startSimulation(SimulatorItem* simulator, bool doReset)
{
    if(simulator->isRunning()){
        //simulator->selectMotionItems();
        TimeBar::instance()->startPlaybackFromFillLevel();

    } else {
        simulator->startSimulation(doReset);
    }
}


void SimulationBar::onStopSimulationClicked()
{
    forEachSimulator(bind(&SimulationBar::stopSimulation, this, _1));

    TimeBar* timeBar = TimeBar::instance();
    if(timeBar->isDoingPlayback()){
        timeBar->stopPlayback();
    }
}


void SimulationBar::stopSimulation(SimulatorItem* simulator)
{
    simulator->stopSimulation();
}
