/**
   @author Shin'ichiro Nakaoka
*/

#ifndef CNOID_UTIL_EIGEN_ARCHIVE_H_INCLUDED
#define CNOID_UTIL_EIGEN_ARCHIVE_H_INCLUDED

#include "ValueTree.h"
#include "EigenUtil.h"
#include <boost/function.hpp>

namespace cnoid {

    template<typename Derived>
    bool readMatrix(const Mapping& mapping, const std::string& key, Eigen::MatrixBase<Derived>& x)
    {
        const Listing& s = *mapping.findListing(key);
        if(s.isValid()){
            const int nr = x.rows();
            const int nc = x.cols();
            const int n = s.size();
            int index = 0;
            if(n > 0){
                for(int i=0; i < nr; ++i){
                    for(int j=0; j < nc; ++j){
                        x(i, j) = s[index++].toDouble();
                        if(index == n){
                            break;
                        }
                    }
                }
            }
            return (index == nr * nc);
        }
        return false;
    }


    template<typename MatrixType>
    bool readMatrix(const Mapping& mapping, const std::string& key,
              boost::function<void(MatrixType&)> setterFunc) {
        MatrixType x;
        if(readMatrix(mapping, key, x)){
            setterFunc(x);
            return true;
        }
        return false;
    }


    template<typename Derived>
    inline void readMatrixEx(const Mapping& mapping, const std::string& key, Eigen::MatrixBase<Derived>& x)
    {
        if(!readMatrix(mapping, key, x)){
            mapping.throwKeyNotFoundException(key);
        }
    }

    template<typename Derived>
    Listing& writeMatrix(Mapping& mapping, const std::string& key, const Eigen::MatrixBase<Derived>& x)
    {
        Listing& s = *mapping.createFlowStyleListing(key);
        const int nr = x.rows();
        const int nc = x.cols();
        if(nc == 1){
            for(int i=0; i < nr; ++i){
                s.append(x(i));
            }
        } else {
            for(int i=0; i < nr; ++i){
                s.appendLF();
                for(int j=0; j < nc; ++j){
                    s.append(x(i, j));
                }
            }
        }
        return s;
    }

    inline bool read(const Mapping& mapping, const std::string& key, Vector3& x)
    {
        return readMatrix(mapping, key, x);
    }

    inline bool read(const Mapping& mapping, const std::string& key, Matrix3& x)
    {
        return readMatrix(mapping, key, x);
    }
    
    inline bool read(const Mapping& mapping, const std::string& key, boost::function<void(Vector3&)> setterFunc)
    {
        return readMatrix<Vector3>(mapping, key, setterFunc);
    }

    inline Listing& write(Mapping& mapping, const std::string& key, const Vector3& x)
    {
        return writeMatrix(mapping, key, x);
    }

    inline Listing& write(Mapping& mapping, const std::string& key, const Matrix3& x)
    {
        return writeMatrix(mapping, key, x);
    }
}

#endif
